"""
Adapted from the original NMpathAnalysis package,
https://github.com/ZuckermanLab/NMpathAnalysis
"""
import numpy as np
from copy import deepcopy
import networkx as nx
from math import log

from msm_we.utils import Interval
from msm_we.utils import get_shape, weighted_choice
from msm_we.utils import reverse_sort_lists
from msm_we.fpt import DirectFPT, NonMarkovFPT


class Ensemble:
    """ Define a base class to store a list of space-continuous trajectories

     e.g.: [ trajectory1, trajectory2,...], each trajectory is just a matrix
    where each row is a snapshot, and each column represents the evolution
    of the corresponding variable.
    """

    def __init__(self, trajectories=None, verbose=False, dtype="float32", discrete=False, lag_time=1, **kwargs):
        """Initialize an object of Ensemble class"""
        super().__init__(**kwargs)
        self.dtype = dtype
        self.discrete = discrete
        self.verbose = verbose
        self._lag_time = lag_time

        if (trajectories is None) or (trajectories == []):
            self.trajectories = []
            self.n_variables = 0
            if verbose:
                print("\nEmpty ensemble generated")

        else:
            # we have a list/array of trajectories
            _n_snapshots, _n_variables = get_shape(trajectories[0])

            traj_length = 0.0  # Average trajectory length
            for element in trajectories:
                traj_length += len(element)

                n_snapshots, n_variables = get_shape(element)

                if n_variables != _n_variables:
                    raise Exception("Error: All the trajectories must have the same " "number of variables")

            self.n_variables = _n_variables
            self.trajectories = trajectories
            n_trajs = len(trajectories)

            traj_length /= n_trajs  # Average

            if verbose:
                print("Read {} ({}-dimensional) trajectories of " "average length {}.".format(n_trajs, n_variables, traj_length))

    def add_trajectory(self, trajectory):
        """Add a single trajectory to the ensemble"""
        if not isinstance(trajectory, np.ndarray):
            trajectory = np.array(trajectory, dtype=self.dtype)

        _n_snapshots, _n_variables = get_shape(trajectory)

        if self.n_variables == 0:  # Empty ensemble
            self.trajectories = [trajectory]
            self.n_variables = _n_variables
            if self.verbose:
                print(self)
        else:
            if self.n_variables != _n_variables:
                raise Exception("All the trajectories in the same ensemble must " "have the same number of variables")
            else:
                self.trajectories.append(trajectory)
                if self.verbose:
                    print(self)

    def __len__(self):
        # returns the number of trajectories in the ensemble
        return len(self.trajectories)

    def __str__(self):
        if self.discrete:
            feature = "Discrete, "
        else:
            feature = "Continuous, "
        return (
            "\n"
            + feature
            + "{} with {} ({}-dimensional) trajectories".format(self.__class__.__name__, self.__len__(), self.n_variables)
            + "\nTotal number of snapshots: {}".format(sum([len(traj) for traj in self]))
        )

    def __add__(self, other):
        ensemble_sum = deepcopy(self)

        for traj in other.trajectories:
            ensemble_sum.add_trajectory(traj)

        return ensemble_sum

    def __iadd__(self, other):
        return self.__add__(other)

    def __iter__(self):
        return iter(self.trajectories)

    def __getitem__(self, arg):
        return self.trajectories[arg]

    def empirical_mfpts(self, stateA, stateB):
        """Calculate mean first passage time using direct counting method"""
        return DirectFPT.mean_fpts(
            self.trajectories, stateA, stateB, discrete=self.discrete, n_variables=self.n_variables, lag_time=self._lag_time,
        )

    def _count_matrix(self, n_states=None, map_function=None):
        """Built the count matrix"""
        if (map_function is None) or (n_states is None):
            raise Exception("The number of states and a map function " "have to be given as argument")

        count_matrix = np.zeros((n_states, n_states))

        for traj in self.trajectories:
            previous_state = "Unknown"
            for snapshot in traj:
                current_state = map_function(snapshot)
                if previous_state != "Unknown":
                    count_matrix[previous_state, current_state] += 1.0
                previous_state = current_state

        return count_matrix

    def _mle_transition_matrix(self, n_states, map_function):
        """Generate transition matrix from count matrix"""
        count_matrix = self._count_matrix(n_states, map_function)
        transition_matrix = count_matrix.copy()

        # Transforming the count matrix to a transition matrix
        for i in range(n_states):
            row_sum = sum(transition_matrix[i, :])
            if row_sum != 0.0:
                transition_matrix[i, :] = transition_matrix[i, :] / row_sum

        return transition_matrix

    def empirical_corr_function(self, stateA, stateB, times, symmetric=True):
        """Calculate correlation function for trajectories"""
        n_dim = self.n_variables

        stateA = Interval(stateA, n_dim) if not self.discrete else stateA
        stateB = Interval(stateB, n_dim) if not self.discrete else stateB

        corr_values = []

        for delay in times:
            assert type(delay) == int
            assert delay >= 1
            sum_ = 0
            counts = 0

            for traj in self.trajectories:
                for i in range(len(traj) - delay):
                    sum_ += (traj[i] in stateA) * (traj[i + delay] in stateB)
                    counts += 1

                    if symmetric:
                        sum_ += (traj[i] in stateB) * (traj[i + delay] in stateA)
                        counts += 1
            corr_values.append(sum_ / counts)

        return corr_values


class PathEnsemble(Ensemble):
    """ Derive a class to store trajectories for path analysis"""

    def __init__(
        self, trajectories=None, verbose=False, dtype="float32", discrete=False, lag_time=1, stateA=None, stateB=None, **kwargs
    ):
        super().__init__(trajectories, verbose, dtype, discrete, lag_time, **kwargs)

        if (stateA is None) or (stateB is None):
            raise Exception(
                "The initial state (stateA) and final state (stateB) \
                 have to be specified"
            )
        self.stateA = stateA
        self.stateB = stateB

    @classmethod
    def from_ensemble(
        cls, ensemble, stateA=None, stateB=None, map_function=None, discrete=False, dtype="float32",
    ):
        """Build a PathEnsemble from an ensemble object or a set of trajectories"""

        list_of_pathsAB = []

        if np.size(ensemble[0][0]):
            n_variables = np.size(ensemble[0][0])
        else:
            n_variables = 1

        if (stateA is None) or (stateB is None):
            raise Exception(
                "The initial state (stateA) and final state (stateB) \
                have to be specified"
            )

        for traj in ensemble.trajectories:
            previous_color = "Unknown"
            pathAB = []
            for _snapshot in traj:

                if map_function is not None:
                    snapshot = map_function(_snapshot)
                else:
                    snapshot = _snapshot

                # color determination
                if not discrete:
                    if snapshot in Interval(stateA, n_variables):
                        color = "A"
                    elif snapshot in Interval(stateB, n_variables):
                        color = "B"
                    else:
                        color = previous_color
                else:
                    if snapshot in stateA:
                        color = "A"
                    elif snapshot in stateB:
                        color = "B"
                    else:
                        color = previous_color

                if color == "A":
                    pathAB.append(snapshot)
                elif (color == "B") and (previous_color == "A"):
                    pathAB.append(snapshot)
                    list_of_pathsAB.append(np.array(pathAB, dtype=dtype))
                    pathAB = []

                previous_color = color

        return cls(list_of_pathsAB, stateA=stateA, stateB=stateB, dtype=dtype, discrete=discrete,)

    def cluster(self, distance_metric, n_cluster=10, method="K-means"):
        raise NotImplementedError("Not implemented yet")


class DiscreteEnsemble(Ensemble):
    """Define a base class to store a list of space-discrete trajectories"""

    def __init__(self, trajectories=None, verbose=False, dtype="int32", discrete=True, lag_time=1, **kwargs):
        """Initialize an object of Discrete Ensemble"""
        super().__init__(trajectories, verbose, dtype, discrete, lag_time, **kwargs)
        if (self.n_variables != 1) and (self.n_variables != 0):
            raise Exception(
                "A discrete trajectory must have a one-dimensional \
                index/variable unless is empty"
            )
        self.n_variables = 1  # by definition

    @classmethod
    def from_ensemble(cls, ens, map_function=None, dtype="int32"):
        """Build a DiscreteEnsemble from an ensemble object or a set of trajectories"""
        # TODO: Check the case when ens is a list of trajs
        if map_function is None:
            raise ValueError("A map function has to be given as argument")

        discrete_trajs_list = []

        if isinstance(ens, Ensemble):
            # it is an Ensemble object
            for traj in ens.trajectories:
                d_traj = np.array([], dtype=dtype)
                for snapshot in traj:
                    d_traj = np.append(d_traj, np.array([map_function(snapshot)]), axis=0)
                discrete_trajs_list.append(d_traj)
            return cls(discrete_trajs_list)
        else:
            # it is a list of trajectories
            d_traj = []
            for traj in ens:
                d_traj += [map_function(snapshot)]
            d_traj = np.array(d_traj, dtype=dtype)

            return cls([d_traj])

    @classmethod
    def from_transition_matrix(cls, transition_matrix, sim_length=None, initial_state=0):
        """Generates a DiscreteEnsemble from the transition matrix"""
        if sim_length is None:
            raise Exception("The simulation length must be given")

        if not isinstance(transition_matrix, np.ndarray):
            transition_matrix = np.array(transition_matrix)

        n_states = len(transition_matrix)
        assert n_states == len(transition_matrix[0])

        current_state = initial_state
        discrete_traj = [initial_state]

        for i in range(sim_length):
            next_state = weighted_choice([k for k in range(n_states)], transition_matrix[current_state, :])
            discrete_traj.append(next_state)
            current_state = next_state

        return cls([np.array(discrete_traj)])


class DiscretePathEnsemble(PathEnsemble, DiscreteEnsemble):
    """Derive a class to store a list of discrete trajectories for path analysis"""

    def __init__(
        self, trajectories=None, verbose=False, dtype="int32", discrete=True, lag_time=1, stateA=None, stateB=None, **kwargs
    ):

        super().__init__(trajectories, verbose, dtype, discrete, lag_time, stateA, stateB, **kwargs)

    @classmethod
    def from_transition_matrix(
        cls, transition_matrix, stateA=None, stateB=None, n_paths=1000, ini_pops=None, max_iters=1000000000,
    ):
        """
        Construct a path ensemble from a transition matrix

        stateA:     list, intitial state

        stateB:     list, final state

        n_paths:    integer, with default value of 1000
                    number of paths to generate

        ini_pops:   list or label, probability distribution over the
                    initial state used to generate the path

                    possible values:
                    ----------------
                    a) None
                    Use a uniform distribution over the states in stateA
                    c) list
                    A list with the explicit values of the populations in stateA
                    that should be used to generate the ensemble
        max_iters:  integer, with default value of 1000000000
                    maximum of iterations for generating path trajectories
        """

        if ini_pops is None:
            ini_pops = [1 / float(len(stateA)) for i in range(len(stateA))]
        elif ini_pops == "ss":
            raise NotImplementedError("Sorry: not yet implemented")

        n_states = len(transition_matrix)
        assert n_states == len(transition_matrix[0])

        d_trajectories = []

        for i in range(n_paths):
            current_state = weighted_choice(stateA, ini_pops)  # Initial state
            path = [current_state]

            for j in range(max_iters):
                next_state = weighted_choice([k for k in range(n_states)], transition_matrix[current_state, :])
                path += [next_state]
                current_state = next_state
                if j + 1 == max_iters:
                    print("\nWARNING: max iteration reached when generating " "the path ensemble, consider to increase max_iters")
                if current_state in stateB:
                    break

            path = np.array(path)
            d_trajectories.append(path)

        return cls(d_trajectories, stateA=stateA, stateB=stateB)

    @classmethod
    def from_ensemble(cls, ensemble, stateA, stateB, map_function=None):
        """Build a DiscreteEnsemble from an ensemble object or a set of trajectories"""
        ens = PathEnsemble.from_ensemble(ensemble, stateA, stateB, map_function, discrete=True, dtype="int32")
        return cls(ens.trajectories, stateA=stateA, stateB=stateB)

    def nm_mfpt(self, ini_probs=None, n_states=None):
        """Compute the mean-first passage time from the transition matrix"""
        t_matrix = self._mle_transition_matrix(n_states)
        ini_state = list(self.stateA)
        final_state = sorted(list(self.stateB))

        return NonMarkovFPT.directional_mfpt(t_matrix, ini_state, final_state, ini_probs)

    def _fundamental_sequences(self, transition_matrix, symmetric=True):
        """Divide/classify the path ensemble into fundamental sequences"""

        fundamental_seqs = []

        for path in self.trajectories:
            if symmetric:
                cmatrix = self._connectivity_matrix(path, transition_matrix * transition_matrix.T)
            else:
                cmatrix = self._connectivity_matrix(path, transition_matrix)

            path_graph = self._graph_from_matrix(cmatrix)
            shortest_path = nx.dijkstra_path(path_graph, path[0], path[-1], "distance")

            fundamental_seqs.append(shortest_path)

        return fundamental_seqs

    def weighted_fundamental_sequences(self, transition_matrix=None, symmetric=True):
        """Generate sorted fundamental sequences with weights"""
        fs_list = self._fundamental_sequences(transition_matrix, symmetric)
        element_count = {}
        tot_count = 0
        for element in fs_list:
            pseudo_index = tuple(element)
            tot_count += 1
            if pseudo_index not in element_count:
                element_count[pseudo_index] = 1
            else:
                element_count[pseudo_index] += 1

        weights = []
        new_fs_list = []
        for key, value in element_count.items():
            new_fs_list.append(key)
            weights.append(value / float(tot_count))

        reversed_sorted_weights, reversed_sorted_new_fs_list = reverse_sort_lists(weights, new_fs_list)

        return reversed_sorted_new_fs_list, reversed_sorted_weights, tot_count

    @staticmethod
    def _graph_from_matrix(matrix):
        """Builds a directed Graph from a matrix like a transition matrix"""

        size = len(matrix)
        assert size == len(matrix[0])
        matrix = np.array(matrix)

        G = nx.DiGraph()

        for node in range(size):
            G.add_node(node)

        for i in range(size):
            for j in range(size):
                if (i != j) and (matrix[i, j] != 0.0):
                    G.add_edge(i, j, distance=-log(matrix[i, j]))
        return G

    @staticmethod
    def _connectivity_matrix(path, matrix):
        """From a given path and a matrix construct a connectivity matrix

        whose elements ij are zero if the transition i->j
        is not observed in the path or (i=j), while keep the rest of the
        elements in the input matrix.

        This way, from the connectivity matrix we could later create a graph
        that represents the path, being the "distance" between nodes equal
        to -log(Tij)

        Tij --> i,j element in the transition matrix

        the path must be 1D array of indexes
        """
        matrix = np.array(matrix)
        path = np.array(path, dtype="int32")

        n_states = len(matrix)
        assert n_states == len(matrix[0])

        c_matrix = np.zeros((n_states, n_states))

        for i in range(len(path) - 1):
            c_matrix[path[i], path[i + 1]] = matrix[path[i], path[i + 1]]

        return c_matrix



import numpy as np

import logging

import westpa.core.binning
from rich.logging import RichHandler

from . import msm_we
from sklearn.cluster import KMeans
import pickle
import base64

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)
log.addHandler(RichHandler())
log.propagate = False


def solve_discrepancy(tmatrix, pi, B):
    """
    Given a transition matrix, solves for the discrepancy function.

    The Poisson equation for the discrepancy function is
    .. math::
        (I - K)h = 1_B - \pi(B), \:\: h \cdot \pi = 0

    however, since :math:`I-K` is singular, we instead solve
    .. math::
        (I - K + \pi \pi^T / || \pi ||^2_2)h = 1_B - \pi(B), \:\: h \cdot \pi = 0
    where :math:`h` is a volumn vector, `1_B` is an indicator function which is 1 in B and 0 everywhere
    else, :math:`\pi` is the steady-state solution of :math:`K`, and `\pi(B)` is a column vector with
    the steady-state value of :math:`\pi(B)` in every element.

    Parameters
    ----------
    tmatrix, 2D array-like: Transition matrix
    pi, array-like: Steady-state distribution for the input transition matrix
    B, array-like: Indices of target states B

    Returns
    --------
    (discrepancy, variance)
    """

    log.info("Computing pi matrix")
    norm = np.dot(pi, pi.T)
    pi_matrix = pi @ pi.T / norm

    b_indicator = np.zeros_like(pi)
    b_indicator[B] = 1.0

    pi_b = np.ones_like(pi)
    pi_b[:] = sum(pi[B])

    discrepancy = np.linalg.solve(np.identity(tmatrix.shape[0]) - tmatrix + pi_matrix,
                                  b_indicator - pi_b)

    variance = np.sqrt(
        np.dot(tmatrix, discrepancy**2) -
        np.dot(tmatrix, discrepancy)**2
    )

    return discrepancy, variance


def get_uniform_mfpt_bins(variance, discrepancy, steady_state, n_desired_we_bins):
    """
    Implements the MFPT-binning strategy described in [1], where bins are groups of microstates that are uniformly
    spaced in the integral of pi * v

    Parameters
    ----------
    variance, array-like: Variance function
    steady_state, array-like: Steady-state distribution
    n_bins int: Number of macrobins

    Returns
    -------

    References
    ----------
    [1] Aristoff, D., Copperman, J., Simpson, G., Webber, R. J. & Zuckerman, D. M.
    Weighted ensemble: Recent mathematical developments. Arxiv (2022).

    """

    # The last two elements of this are the basis and target states respectively
    pi_v = steady_state * variance
    pi_v_sort = np.argsort(discrepancy).squeeze()
    cumsum = np.cumsum(pi_v[pi_v_sort])

    spacing = sum(pi_v) / n_desired_we_bins

    bin_states = np.full_like(steady_state, fill_value=np.nan)
    for i in range(n_desired_we_bins):
        lower, upper = spacing * i, spacing * (i+1)
        log.debug(f"Checking for states with pi_v between {lower}, {upper}")

        indices_in_bin = np.argwhere(
            (lower < cumsum) & (cumsum <= upper)
        )
        states_in_bin = pi_v_sort[indices_in_bin]

        log.info(f"Found that bin {i} contains microstates {states_in_bin}")
        bin_states[states_in_bin] = i

    return bin_states


def get_clustered_mfpt_bins(variance, steady_state, n_desired_we_bins):
    """
    Implements the MFPT-binning strategy described in [1], where bins are groups of microstates that are uniformly
    spaced in the integral of pi * v

    Parameters
    ----------
    variance, array-like: Variance function
    steady_state, array-like: Steady-state distribution
    n_bins int: Number of macrobins

    Returns
    -------

    References
    ----------
    [1] Aristoff, D., Copperman, J., Simpson, G., Webber, R. J. & Zuckerman, D. M.
    Weighted ensemble: Recent mathematical developments. Arxiv (2022).

    """

    # The last two elements of this are the basis and target states respectively
    pi_v = steady_state * variance

    clusterer = KMeans(n_clusters=n_desired_we_bins)

    # -2 so you don't cluster the basis/target states
    # TODO: Handle those in a more general way, though from msm_we you're guaranteed they'll be the last two states
    # we_bin_assignments = clusterer.fit_predict(pi_v.reshape(-1, 1))
    we_bin_assignments = clusterer.fit_predict(pi_v[:-2].reshape(-1, 1))

    bin_states = np.full_like(steady_state, fill_value=np.nan)
    for i in range(n_desired_we_bins):
        states_in_bin = np.argwhere(we_bin_assignments == i).squeeze()
        bin_states[states_in_bin] = i
        log.info(f"Found that bin {i} contains microstates {states_in_bin}")

    return bin_states


class OptimizedBinMapper(westpa.core.binning.FuncBinMapper):

    def __init__(self, *args, **kwargs):

        if len(args) + len(kwargs) == 1:
            log.info("Single argument provided to binmapper initializer, attempting to unpickle")
            self.unpickle(kwargs['bytestring'])

        elif len(args) + len(kwargs) > 1:
            log.info("Multiple arguments provided to binmapper initializer, creating new object")
            self.create_new(*args, **kwargs)

    def unpickle(self, bytestring):

        # TODO: This may need to be provided as a b64-encoded bytestring, rather than plain bytes,
        #  in which case we need to quickly encode it back to bytes. This can be done with
        #   pickle.loads(base64.b64decode(collection.metadata['bin_mapper']))

        loaded = pickle.loads(base64.b64decode(bytestring))

        assert type(loaded) is type(self), "Invalid pickled object provided"

        # TODO: Probably just update the attributes I explicitly care about, this seems a little insecure
        for k, v in vars(loaded).items():
            setattr(self, k, v)

    def create_new(self,
                 nbins: int,
                 n_original_pcoord_dims: int,
                 target_pcoord_bounds,
                 basis_pcoord_bounds,
                 previous_binmapper,
                 microstate_mapper: dict,
                 stratified_clusterer: msm_we.StratifiedClusters,
                 *args,
                 **kwargs
                 ):
        """
        Creates an OptimizedBinMapper, suitable for use with the optimization workflow

        Parameters
        ----------
        nbins: int, Number of WE bins
        n_original_pcoord_dims: int, Number of dimensions in the original user-supplied progress coordinate
        microstate_mapper: dict, Mapping of microstates to WE bins
        """

        super().__init__(func=self.mapper, nbins=nbins, args=args, kwargs=kwargs)

        self.microstate_mapper = microstate_mapper
        self.n_original_pcoord_dims = n_original_pcoord_dims

        self.base_mapper = previous_binmapper

        self.simple_model = msm_we.modelWE()
        self.simple_model.pcoord_ndim = n_original_pcoord_dims
        self.simple_model.basis_pcoord_bounds = basis_pcoord_bounds
        self.simple_model.target_pcoord_bounds = target_pcoord_bounds

        self.clusterer = stratified_clusterer
        self.clusterer.model = self.simple_model

        self.clusterer.model.n_clusters = 2
        for cluster_model in self.clusterer.cluster_models:
            if hasattr(cluster_model, 'cluster_centers_'):
                self.clusterer.model.n_clusters += len(cluster_model.cluster_centers_)
        log.info(f"Clusterer has {self.clusterer.model.n_clusters} total clusters")
        log.info(f"Clusterer remap is {self.clusterer.we_remap}")

    def mapper(self, coords, mask, output, *args, **kwargs):

        # TODO: When analyzing augmented coordinates, this needs to pull just the child pcoord -- but when analyzing
        #   non-augmented, that won't exist
        # final_coords = coords[:,-1]
        if len(coords.shape) == 3:
            final_coords = coords[:, -1]
        else:
            final_coords = coords

        log.debug(f"Mapping pcoords {final_coords}")

        # To use stratified clustering, first load the ORIGINAL pcoords into stratified.pcoord1List, then call
        #   stratified.predict().
        # Segments in the basis will be assigned to whatever n_clusters is, and targets will be assigned to whatever
        #   n_clusters + 1 is.
        #  This isn't actually used for anything else, and no clustering happens for these, so I can actually
        #  set these arbitrarily.
        original_pcoords = final_coords[:, :self.n_original_pcoord_dims]

        basis_we_bin_idx, target_we_bin_idx = self.nbins-2, self.nbins-1

        log.debug(f"Original pcoords dimensionality was {self.n_original_pcoord_dims}")
        log.debug(f"Original pcoords had shape {original_pcoords.shape}")
        self.clusterer.model.pcoord1List = original_pcoords

        base_bins = self.base_mapper.assign(original_pcoords)
        log.debug(f"Base bin mapper mapped to {base_bins}")

        # Now, do stratified clustering on the rest of the coordinates.
        # Each segment will be
        #   1. Assigned a WE bin, based on its pcoords and the provided bin mapper
        #   2. Discretized, according to the k-means model associated with that WE bin
        log.debug(f"About to cluster coords of shape {final_coords.shape}")
        stratified_cluster_assignments = self.clusterer.predict(final_coords)

        log.debug(f"Got microstate assignments {stratified_cluster_assignments}")

        # TODO: Map microstates to new WE bins, and populate we_bin_assignments
        # I have a microstate for each segment now -- I need to refer to my mapping of microstates to WE bins, which
        #   just comes from my optimization step
        log.debug(f"Mapping microstates to WE bins using {self.microstate_mapper}")

        we_bin_assignments = np.array([float(self.microstate_mapper[microstate])
                                       if microstate < len(self.microstate_mapper) else -1
                                       for microstate in stratified_cluster_assignments
                                      ])

        log.debug(f"Basis WE bin is labeled {basis_we_bin_idx}, target WE bin is labeled {target_we_bin_idx}")
        log.debug(f"WE bin assignments before correcting basis/target are {we_bin_assignments}")

        we_bin_assignments[self.clusterer.model.is_WE_target(final_coords)] = target_we_bin_idx
        we_bin_assignments[self.clusterer.model.is_WE_basis(final_coords)] = basis_we_bin_idx

        zipped_assignments = np.array(list(zip(original_pcoords.reshape(-1), we_bin_assignments)))
        zip_sort = np.argsort(original_pcoords.reshape(-1))

        log.debug(f"WE bin assignments are {zipped_assignments[zip_sort]}")

        for i in range(len(output)):
            output[i] = we_bin_assignments[i]

        assert not np.isnan(output).any()

        return output.astype(int)



"""
Adapted from the original NMpathAnalysis package,
https://github.com/ZuckermanLab/NMpathAnalysis
"""
from copy import deepcopy
import numpy as np
import operator


class Interval:
    """Intervals are in general defined as half-open interval [start,end).

    in any case, in each dimension the interval is specified using a list [a,b]
    where a < b

    - For 1D (single) interval a single list in the form [a,b] has to be given

    - The union of multiple (1D) intervals can be specified as:
        [[a,b],[c,d],...]

    - A list of lists [[a,b],[c,d],...] are used for a n-dimensional
        intervals, one for each dimension, i.e, len(interval) = n_variables

    - A list-of-lists-of-lists for the mathematical union of n-dimensional
        intervals'

        [ [[a,b],[c,d],...],  [[e,f],[g,h],...], ... ]

    """

    def __init__(self, interval_set, n_variables):
        self.interval_set = interval_set
        self.n_variables = n_variables

    def __contains__(self, item):

        shape = np.array(self.interval_set).shape

        len_shape = len(shape)

        if (self.n_variables == 1) and (len_shape == 1):  # single 1D interval
            return self.interval_set[0] <= item < self.interval_set[1]

        elif (self.n_variables == 1) and (len_shape == 2):  # union of multiple 1D intervals
            return any([(item in Interval(self.interval_set[i], 1)) for i in range(shape[0])])

        elif (self.n_variables > 1) and len_shape == 2:  # n-dimensional interval
            return all([(item[i] in Interval(self.interval_set[i], 1)) for i in range(shape[0])])

        elif len(shape) == 3:  # union of n-dimensional intervals
            return any([(item in Interval(self.interval_set[i], self.n_variables)) for i in range(shape[0])])
        else:
            raise Exception("The given interval has not the expected shape")


def reverse_sort_lists(list_1, list_2):
    """Reverse sorting two list based on the first one"""
    list_1_sorted, list_2_sorted = zip(*sorted(zip(list_1, list_2), key=operator.itemgetter(0), reverse=True))
    return list_1_sorted, list_2_sorted


def weighted_choice(list_, weights=None):
    """Select an element from a list with probability from weights"""
    size = len(list_)
    if weights is not None:
        assert size == len(weights)

    if weights is None:
        probs = np.array([1 / float(size) for i in range(size)])
    else:
        probs = np.array(weights) / sum(weights)  # just in case

    rand = np.random.random()

    _sum = 0
    for i in range(size):
        if _sum <= rand < _sum + probs[i]:
            choice = i
            break
        else:
            _sum += probs[i]

    return list_[choice]


def get_shape(trajectory):
    """Get the shape of a trajectory array in tuple (n_snapshots, n_variables)"""
    shape = np.array(trajectory).shape

    if len(shape) == 1:
        n_snapshots = shape[0]
        n_variables = 1
        if n_variables == 0:
            raise Exception(
                "The shape {} of the trajectory/array \
            given is not as expected".format(
                    shape
                )
            )
    elif len(shape) == 2:
        n_snapshots = shape[0]
        n_variables = shape[1]
    else:
        raise Exception(
            "The shape {} of the trajectory/array given is not \
            as expected".format(
                shape
            )
        )

    return n_snapshots, n_variables


def num_of_nonzero_elements(my_vector):
    """Returns the number of non-zero elements in a vector"""
    counter = 0
    for element in my_vector:
        if element != 0:
            counter += 1
    return counter


def normalize_markov_matrix(transition_matrix, reversible=False):
    """Transform a matrix of positive elements to a markov-like matrix

    by dividing each row by the sum of the elements of the row.
    """
    t_matrix = np.array(transition_matrix, dtype=np.float64)
    if reversible:
        t_matrix = t_matrix.T + t_matrix

    n_states = len(t_matrix)
    assert n_states == len(t_matrix[0])

    for i in range(n_states):
        if (t_matrix[i, :] < 0).any():
            raise ValueError(
                "All the elements in the input \
            matrix must be non-negative"
            )
        t_matrix[i, :] = normalize(t_matrix[i, :])

    return t_matrix


def normalize(my_vector):
    """Normalize a vector

    by dividing each element by the total sum of all its elements
    """
    my_vector = np.array(my_vector)
    size = len(my_vector)

    sum_ = sum(my_vector)
    if sum_ != 0.0:
        for i in range(size):
            my_vector[i] = my_vector[i] / sum_
    return my_vector


def random_markov_matrix(n_states=5, seed=None):
    """Returns a random transition markov matrix"""
    if seed is not None:
        np.random.seed(seed)
    t_matrix = np.random.random((n_states, n_states))
    return normalize_markov_matrix(t_matrix)


def check_tmatrix(t_matrix, accept_null_rows=True):
    """Check if the given matrix is actually a row-stochastic transition matrix

     i.e, all the elements are non-negative and the rows add to one.
    If the keyword argument accept_null_rows is True, is going
    to accept rows where all the elements are zero. Those "problematic"
    states are going to be removed later if necessary by clean_tmatrix.
    """

    def value_error():
        raise ValueError("The object given is not a transition matrix")

    n_states = len(t_matrix)
    if not (n_states == len(t_matrix[0])):
        value_error()

    for index, row in enumerate(t_matrix):
        sum_ = 0.0
        for element in row:
            if element < 0.0:
                value_error()
            sum_ += element

        if accept_null_rows:
            if not (np.isclose(sum_, 1.0, atol=1e-6) or sum_ == 0.0):
                value_error()
        else:
            if not np.isclose(sum_, 1.0, atol=1e-6):
                value_error()

    return False


def clean_tmatrix(transition_matrix, rm_absorbing=True):
    """Removes the states/indexes with no transitions and that are absorbing

    if the the keyword argument rm_absorbing is true
    Returns the "clean" transition matrix and a list with the
    removed states/indexes (clean_tmatrix, removed_states)
    """
    t_matrix = deepcopy(transition_matrix)
    n_states = len(transition_matrix)

    # Removing the non-visited states and absorbing states
    removed_states = []
    for index in range(n_states - 1, -1, -1):
        if not any(t_matrix[index]):  # non-visited
            t_matrix = np.delete(t_matrix, index, axis=1)
            t_matrix = np.delete(t_matrix, index, axis=0)
            removed_states.append(index)
        elif t_matrix[index, index] == 1.0:  # absorbing state
            if not all([t_matrix[index, j] == 0.0 for j in range(n_states) if j != index]):
                raise ValueError(
                    "The sum of the elements in a row of the \
                    transition matrix must be one"
                )
            t_matrix = np.delete(t_matrix, index, axis=1)
            t_matrix = np.delete(t_matrix, index, axis=0)
            removed_states.append(index)

    # Renormalizing just in case
    t_matrix = normalize_markov_matrix(t_matrix)

    return t_matrix, removed_states


def pops_from_tmatrix(transition_matrix):
    """Calculate the eigen values and eigen vectors of the transposed transition matrix

    Parameters
    ----------
    transition_matrix : ndarray with shape = (n_states, n_states)

    Returns
    -------
    the solution, p, of K.T p = p where K.T is the transposed transition matrix
    """

    check_tmatrix(transition_matrix)

    n_states = len(transition_matrix)

    # Cleaning the transition matrix
    cleaned_matrix, removed_states = clean_tmatrix(transition_matrix)

    # Computing
    eig_vals, eig_vecs = np.linalg.eig(cleaned_matrix.T)
    eig_vecs = eig_vecs.T  # for convenience, now every row is an eig_vector

    eig_vals_close_to_one = np.isclose(eig_vals, 1.0, atol=1e-6)
    real_eig_vecs = [not np.iscomplex(row).any() for row in eig_vecs]

    new_n_states = n_states - len(removed_states)

    ss_solution = np.zeros(new_n_states)  # steady-state solution
    for is_close_to_one, is_real, eigv in zip(eig_vals_close_to_one, real_eig_vecs, eig_vecs):
        if (
            is_close_to_one
            and is_real
            and num_of_nonzero_elements(eigv) > num_of_nonzero_elements(ss_solution)
            and ((eigv <= 0).all() or (eigv >= 0).all())
        ):
            ss_solution = eigv

    if (ss_solution == 0.0).all():
        raise Exception(
            "No steady-state solution found for \
        the given transition matrix"
        )

    ss_solution = normalize(ss_solution).real

    # Now we have to insert back in the solution, the missing
    # elements with zero probabilities
    for index in sorted(removed_states):
        ss_solution = np.insert(ss_solution, index, 0.0)

    return ss_solution


def pops_from_nm_tmatrix(transition_matrix):
    """Computes the populations of the real/physical states

    from a non-Markovian transtion matrix with shape (2*n_states, 2*n_states)
    """
    check_tmatrix(transition_matrix, accept_null_rows=True)

    size = len(transition_matrix)

    if size % 2 != 0:
        raise ValueError("The non-Markovian transition matrix has to " "have an even number of columns/rows")

    n_states = size // 2  # Real/physical microstates

    pops_nm = pops_from_tmatrix(transition_matrix)

    pops = np.zeros(n_states)

    for i in range(n_states):
        pops[i] = pops_nm[2 * i] + pops_nm[2 * i + 1]

    return pops


def map_to_integers(sequence, mapping_dict=None):
    """Map a sequence of elements to a sequence of integers
    for intance, maps [1, 'a', 1, 'b', 2.2] to [0, 1, 0, 2, 3]
    """
    if mapping_dict is None:
        mapping_dict = {}

    new_sequence = np.zeros(len(sequence), dtype="int64")

    counter = 0

    for i, element in enumerate(sequence):
        if element not in mapping_dict.keys():
            mapping_dict[element] = counter
            counter += 1

        new_sequence[i] = mapping_dict[element]
    return new_sequence, mapping_dict


def pseudo_nm_tmatrix(markovian_tmatrix, stateA, stateB):
    """Obtain a pseudo non-Markovian transition matrix from a Markovian transiton matrix

    The pseudo Markovian matrix has a shape of (2 n_states, 2 n_states)
    """

    check_tmatrix(markovian_tmatrix)
    n_states = len(markovian_tmatrix)

    # pseudo non-Markovian transition matrix
    p_nm_tmatrix = np.zeros((2 * n_states, 2 * n_states))

    for i in range(2 * n_states):
        for j in range(2 * n_states):
            p_nm_tmatrix[i, j] = markovian_tmatrix[i // 2, j // 2]

    for i in range(n_states):
        for j in range(n_states):
            if (i in stateB) or (j in stateB):
                p_nm_tmatrix[2 * i, 2 * j] = 0.0
            if (i in stateA) or (j in stateA):
                p_nm_tmatrix[2 * i + 1, 2 * j + 1] = 0.0
            if (not (j in stateA)) or (i in stateA):
                p_nm_tmatrix[2 * i + 1, 2 * j] = 0.0
            if (not (j in stateB)) or (i in stateB):
                p_nm_tmatrix[2 * i, 2 * j + 1] = 0.0

    check_tmatrix(p_nm_tmatrix)  # just in case
    return p_nm_tmatrix



"""
Adapted from the original NMpathAnalysis package,
https://github.com/ZuckermanLab/NMpathAnalysis
"""
import numpy as np
from copy import deepcopy
import tqdm.auto as tqdm

import msm_we.utils as utils
from msm_we.utils import Interval


class DirectFPT:
    @classmethod
    def mean_fpts(
        cls,
        trajectories,
        stateA=None,
        stateB=None,
        discrete=True,
        n_variables=None,
        lag_time=1,
    ):
        """Empirical mean first passage times (MFPTs) calculation (no model
        involved) by tracing the trajectories. Notice the difference between
        notation between FPTs and MFPTs.

        Parameters:
        -----------
        trajectories:   List of trajectories [traj1, traj2, traj4], each trajectory
                        can be a one dimensional array, e.g.,
                            [[1,2,1, ...], [0,1,1, ...], ... ]
                        or a multidimensional array (matrix) where each column
                        represents the evolution of a variable.

                        Important: If a single trajectory is given as argument it
                        also has to be inside a list (e.g. [traj1])

        stateA, stateB: List of integers
                        If the trajectories are discrete (discrete = True), both
                        states are a list of indexes. However, if the trajectories
                        are not discrete, the states are "interval" objects
                        (see Interval class).

        lag_time:       integer
                        Lag time used, the trajectory is "observed" every lag_time
                        time steps

        discrete:       boolean
                        False when the trajectories are are not discrete. In that
                        case the macrostates stateA and stateB are considered
                        interval objects.

        n_variables:    integer
                        If the trajectory is space continuous,the number of
                        variables/dimensions is needed. In this case every
                        trajectory inside "trajectories" should have the same
                        number of dimensions.

        Returns
        -------
        A dictionary with the keys: 'mfptAB', 'std_err_mfptAB', 'mfptBA',
        'std_err_mfptBA' and the corresponding values. Those values are already
        multiplied by the lag_time used (not the physical units).
        """

        passage_timesAB, passage_timesBA, tb_values = cls.fpts(
            trajectories, stateA, stateB, discrete, n_variables, lag_time
        )
        n_AB = len(passage_timesAB)
        n_BA = len(passage_timesBA)

        if sum(passage_timesAB):
            mfptAB = float(sum(passage_timesAB)) / n_AB
            std_err_mfptAB = np.std(passage_timesAB) / np.sqrt(n_AB)
        else:
            print("WARNING: No A->B events observed")
            mfptAB = "NaN"
            std_err_mfptAB = "NaN"

        if sum(passage_timesBA):
            mfptBA = float(sum(passage_timesBA)) / n_BA
            std_err_mfptBA = np.std(passage_timesBA) / np.sqrt(n_BA)
        else:
            print("WARNING: No B->A events observed")
            mfptBA = "NaN"
            std_err_mfptBA = "NaN"

        kinetics = {
            "mfptAB": mfptAB,
            "std_err_mfptAB": std_err_mfptAB,
            "mfptBA": mfptBA,
            "std_err_mfptBA": std_err_mfptBA,
        }

        print("Number of A->B/B->A  events: {}/{}".format(n_AB, n_BA))

        return kinetics

    @classmethod
    def fpts(
        cls,
        trajectories,
        stateA=None,
        stateB=None,
        discrete=True,
        n_variables=None,
        lag_time=1,
    ):
        """Empirical first passage times (FPTs) calculation (no model involved)
        by tracing the trajectories. IMPORTANT: Notice the difference in notation
        between FPTs and MFPTs.

        Parameters:
        -----------
        trajectories:   List of trajectories [traj1, traj2, traj4], each trajectory
                        can be a one dimensional array, e.g.,
                            [[1,2,1, ...], [0,1,1, ...], ... ]
                        or a mutidimensional array (matrix) where each column
                        represents the evolution of a variable.

                        Important: If a single trajectory is given as argument it
                        also has to be inside a list (e.g. [traj1])

        stateA, stateB: List of integers
                        If the trajectories are discrete (discrete = True), both
                        states are a list of indexes. However, if the trajectories
                        are not discrete, the states are "interval" objects
                        (see Interval class).

        lag_time:       integer
                        Lag time used, the trajectory is "observed" every lag_time
                        time steps

        discrete:       boolean
                        False when the trajectories are are not discrete. In that
                        case the macrostates stateA and stateB are considered
                        interval objects.

        n_variables:    integer
                        If the trajectory is space continuous,the number of
                        variables/dimensions is needed. In this case every
                        trajectory inside "trajectories" should have the same
                        number of dimensions.

        Returns
        -------
        A tuple of two 1D-ndarray (array1, array2), the first one contains the
        observed first passage times A->B and the second one the FPTs B->A. Those
        values are already multiplied by the lag_time used (not the physical units)
        """

        if (stateA is None) or (stateB is None):
            raise Exception(
                "The final and initial states have " "to be defined to compute the MFPT"
            )

        if not discrete:
            """
            The states are considered/transformed-to intervals if the Ensemble
            is a set of continuous trajectories
            """
            if n_variables is None:
                raise Exception(
                    "In continuous trajectories the number of " "variables is needed"
                )

            stateA = Interval(stateA, n_variables)
            stateB = Interval(stateB, n_variables)

        passage_timesAB = []
        passage_timesBA = []
        tb_values = []

        for traj in trajectories:
            previous_color = "Unknown"
            tb_counter = 0  # event duration counter
            fpt_counter = 0  # first passage time counter
            for i in range(0, len(traj), lag_time):
                snapshot = traj[i]
                tb_counter += 1
                # state and color determination
                if snapshot in stateA:
                    color = "A"
                elif snapshot in stateB:
                    color = "B"
                else:
                    color = previous_color
                    tb_counter += 1

                # passage times
                if (color == "A") or (color == "B"):
                    fpt_counter += 1

                if previous_color == "A" and color == "B":
                    tb_values.append(tb_counter)
                    passage_timesAB.append(fpt_counter)
                    fpt_counter = 0
                elif previous_color == "B" and color == "A":
                    tb_values.append(tb_counter)
                    passage_timesBA.append(fpt_counter)
                    fpt_counter = 0
                elif previous_color == "Unknown" and (color == "A" or color == "B"):
                    fpt_counter = 0

                if (snapshot in stateA) or (snapshot in stateB):
                    tb_counter = 0

                previous_color = color

        passage_timesAB = np.array(passage_timesAB) * lag_time
        passage_timesBA = np.array(passage_timesBA) * lag_time

        return passage_timesAB, passage_timesBA, tb_values


class MatrixFPT:
    """Define a base class for calculating FPTs using transition matrix"""

    @classmethod
    def mean_fpts(cls, tmatrix, stateA, stateB, lag_time=1):
        """Calculting mean-first passave time for at transition matrix

        Derived class will implement this function differently.
        """
        pass

    @classmethod
    def directional_mfpt(
        cls, transition_matrix, stateA, stateB, ini_probs=None, lag_time=1
    ):
        """Computes the mean-first passage time in a single direction using a recursive procedure

        This method is useful when there is no B->A ensemble but only A->B transitions,
        for instance when B is absorbing.

        Parameters:
        -----------
        transition_matrix:  Numpy 2D array

        stateA, stateB:     List of integers
                            Both states are a list of indexes.

        ini_probs:          List of float, default is None
                            initial probabilities in stateA

        lag_time:           integer
                            Lag time used, the trajectory is "observed" every lag_time
                            time steps

        Returns
        -------
        mean-first passage time from A->B


        """
        lenA = len(stateA)
        lenB = len(stateB)

        if ini_probs is None:
            ini_probs = [1.0 / lenA for i in range(lenA)]

        t_matrix = deepcopy(transition_matrix)

        ini_state = list(stateA)
        f_state = sorted(list(stateB))

        assert lenA == len(ini_probs)

        for i in range(lenB - 1, -1, -1):
            t_matrix = np.delete(t_matrix, f_state[i], axis=1)
            t_matrix = np.delete(t_matrix, f_state[i], axis=0)
            for j in range(lenA):
                if f_state[i] < ini_state[j]:
                    ini_state[j] = ini_state[j] - 1

        new_size = len(t_matrix)

        mfptAB = 0.0

        m = np.zeros(new_size)
        idty = np.identity(new_size)
        c = np.array([1.0 for i in range(new_size)])

        m = np.dot(np.linalg.inv(idty - t_matrix), c)

        for i in range(len(ini_state)):
            k = ini_state[i]
            mfptAB += ini_probs[i] * m[k]
        mfptAB = mfptAB / sum(ini_probs)

        return mfptAB * lag_time

    @classmethod
    def mfpts_to_target_microstate(cls, transition_matrix, target, lag_time=1):
        """Computes all the mean-first passage to a target microstate (k)

        Returns a list where the i-element is mfpt(i->k). This function is
        useful to compute the mfpt matrix.

        Parameters:
        -----------
        transition_matrix:  Numpy 2D array

        target:             Integer number that specifies the index of the state. The indexes
                            should be consistent with the transition matrix and python
                            (i.e. starting from 0)

        lag_time:           Integer
                            Lag time used, the trajectory is "observed" every lag_time
                            time steps

        Returns
        -------
        a list where the i-element is mfpt(i->k). This function is
        useful to compute the mfpt matrix.

        """

        t_matrix = deepcopy(transition_matrix)

        t_matrix = np.delete(t_matrix, target, axis=1)
        t_matrix = np.delete(t_matrix, target, axis=0)

        new_size = len(t_matrix)

        m = np.zeros(new_size)
        idty = np.identity(new_size)
        c = np.array([1.0 for i in range(new_size)])

        m = np.dot(np.linalg.inv(idty - t_matrix), c)
        m = np.insert(m, target, 0.0)

        return m * lag_time

    @classmethod
    def mfpts_matrix(cls, transition_matrix, lag_time=1):
        """Calculate MFPT matrix, i.e., the matrix where the ij-element is MFPT(i->j)

        Parameters:
        -----------
        transition_matrix:  Numpy 2D array

        lag_time:           Integer
                            Lag time used, the trajectory is "observed" every lag_time
                            time steps

        Returns
        -------
        mean-first passage time matrix with ij-element of MFPT(i->j)
        """

        size = len(transition_matrix)
        temp_values = []

        for i in range(size):
            temp_values.append(
                cls.mfpts_to_target_microstate(transition_matrix, i, lag_time)
            )

        mfpt_m = np.array(temp_values).T  # to nummpy array and transposed
        return mfpt_m

    @classmethod
    def min_commute_time(cls, matrix_of_mfpts):
        """Calculate minimum commuting time (round trip time) between all pairs

        of microstates from the matrix of mfpts. It also returns the indexes
        of the pair of microstates involved.

        Parameters:
        -----------
        matrix_of_mfpts:    Numpy 2D array
                            matrix of MFPTs with ij-element of MFPT(i->j)

        Returns
        -------
        Minimum commuting time (round trip time) between all pairs
        of microstates and the indexes of the pair of microstates involved.

        """

        matrix_of_mfpts = np.array(matrix_of_mfpts)

        n_states = len(matrix_of_mfpts)
        assert n_states == len(matrix_of_mfpts[0]) and n_states >= 2

        # Initial values, arbitrary choice
        index_i = 0
        index_j = 1

        commute_times = matrix_of_mfpts + matrix_of_mfpts.T
        min_ct = commute_times[index_i, index_j]

        for i in range(n_states):
            for j in range(i + 1, n_states):
                if commute_times[i, j] < min_ct:
                    min_ct = commute_times[i, j]
                    index_i = i
                    index_j = j

        return min_ct, index_i, index_j

    @classmethod
    def max_commute_time(cls, matrix_of_mfpts):
        """Calculate maximum commuting time (round trip time) between all pairs

        of microstates from the matrix of mfpts. It also returns the indexes
        of the pair of microstates involved.

        Parameters:
        -----------
        matrix_of_mfpts:    Numpy 2D array
                            matrix of MFPTs with ij-element of MFPT(i->j)

        Returns
        -------
        Maximum commuting time (round trip time) between all pairs
        of microstates and the indexes of the pair of microstates involved.

        """

        matrix_of_mfpts = np.array(matrix_of_mfpts)

        n_states = len(matrix_of_mfpts)
        assert n_states == len(matrix_of_mfpts[0]) and n_states >= 2

        # Initial values, arbitrary choice
        index_i = 0
        index_j = 1

        commute_times = matrix_of_mfpts + matrix_of_mfpts.T
        max_ct = commute_times[index_i, index_j]

        for i in range(n_states):
            for j in range(i + 1, n_states):
                if commute_times[i, j] > max_ct:
                    max_ct = commute_times[i, j]
                    index_i = i
                    index_j = j

        return max_ct, index_i, index_j

    @classmethod
    def fpt_distribution(
        cls,
        t_matrix,
        initial_state,
        final_state,
        initial_distrib,
        min_power=1,
        max_power=12,
        max_n_lags=100,
        lag_time=1,
        dt=1.0,
        clean_recycling=False,
        logscale=False,
    ):
        """Calculated distribution of first passage times from transition matrix

        Parameters:
        -----------
        t_matrix:           Numpy 2D array

        initial_state,
        final_states:       List of integer numbers
                            Specifies the indexes of initial and final states.

        ini_probs:          List of float, default is None
                            initial probabilities for initial states

        min_power,
        max_power:          Integer
                            The minimum and maximum power when the FPT distribution is
                            shown in logscale such as (10^min_power, 10^max_power)*lag_time*dt.

        max_n_lags:         Integer
                            maximum number of lags when the FPT distribution is shown in linear
                            scale such as (0, max_n_logs)*lag_time*dt. When in logscale, this is number
                            of points to shown in the range of (10^min_power, 10^max_power)*lag_time*dt.

        lag_time:           Integer
                            Lag time used, the trajectory is "observed" every lag_time
                            time steps

        dt:                 Float
                            Time step

        clean_recycling:    Bool
                            Cleaning the recycling of steady state simulation if True

        logscale:           Bool
                            Option to use logscale for FPT time in the distribution


        Returns
        -------
        Distributions of first passage times

        """

        # copy everything since they are going to be modified
        tmatrix = np.copy(t_matrix)

        ini_state = list(initial_state)
        f_state = sorted(list(final_state))

        assert len(ini_state) == len(initial_distrib)
        # Designate target state 0 as the only target state,
        # and adding all the fluxes into the other target states into state 0.
        # Namely reassign any flux into any target state into target state 0.
        tmatrix[:, f_state[0]] = np.sum(tmatrix[:, f_state], axis=1)

        # Remove all other target states
        for i in range(len(f_state) - 1, 0, -1):
            tmatrix = np.delete(tmatrix, f_state[i], axis=1)
            tmatrix = np.delete(tmatrix, f_state[i], axis=0)

            # For each initial state with a greater index than the target state we're cleaning.
            # Decrement the index by 1 to account for the removed, cleaned state.
            for j in range(len(ini_state)):
                if f_state[i] < ini_state[j]:
                    ini_state[j] = ini_state[j] - 1

        # Clean the recycling if necessary.
        # Get rid of recycling boundary conditions, otherwise  we're getting a  CDF
        if clean_recycling:
            tmatrix[f_state, :] = 0.0
            tmatrix[f_state, f_state] = 0.0

        # The new target state is the single state since all other target states have been reassigned to.
        f_state = f_state[0]
        new_n_states = len(tmatrix)
        list_of_pdfs = np.empty((len(ini_state), max_n_lags), dtype=np.float64)
        prevFmatrix = np.empty_like(tmatrix)

        # Option to set the list of lag time in logscale since FPT can be a wide distribution in several orders
        if logscale:
            lag_list = np.logspace(min_power, max_power, max_n_lags, dtype=int)
        else:
            lag_list = np.arange(0, max_n_lags, dtype=int)

        # for each ini_state calculate the FPT distribution from transition matrix
        for istateIndex in range(len(ini_state)):
            prevFmatrix = tmatrix.copy()
            Fmatrix = np.zeros((new_n_states, new_n_states))
            list_of_pdfs[istateIndex, 0] = tmatrix[ini_state[istateIndex], f_state]

            cls.calc_fmatrix(
                Fmatrix,
                tmatrix,
                prevFmatrix,
                list_of_pdfs,
                lag_list,
                ini_state,
                istateIndex,
                f_state,
            )

        # Nomalize the FPT distribution and output
        sum_ = np.sum(initial_distrib)
        initial_distrib = np.array(initial_distrib)

        density = np.sum(initial_distrib[:, None] * list_of_pdfs, axis=0) / sum_

        dt2 = lag_time * dt
        if logscale:
            # For logscale the dts at different t are different, we need to let FPT(t)
            # absorb them. Otherwise we have to use dt in variable size to calculate mean
            # value such as integration of t*dt*FPT(t).
            dens_list = [[0, 0]] + [[lag_list[0] * dt2, density[0] * lag_list[0] / dt2]]
            for i in range(1, len(lag_list)):
                dens_list += [
                    [
                        lag_list[i] * dt2,
                        density[i] * (lag_list[i] - lag_list[i - 1]) / dt2,
                    ]
                ]
            density_vs_t = np.array(dens_list)
        else:
            density_vs_t = np.array(
                [[0, 0]]
                + [[(i + 1) * dt2, dens / dt2] for i, dens in zip(lag_list, density)]
            )
        # normalized to 1
        density_vs_t[:, 1] /= sum(density_vs_t[:, 1])
        return density_vs_t

    @staticmethod
    def adaptive_fpt_distribution(
        Tmatrix,
        initial_states,
        initial_state_probs,
        target_states,
        tau=1,
        increment=5,
        fine_increment=1.2,
        relevant_thresh=1e-4,
        max_steps=int(1e6),
        max_time=np.inf,
        explicit_renormalization=False,
        verbose=False,
    ):
        """
        Adaptively computes a first-passage time distribution.

        Starting at t=tau, compute the probability flowing into the target at t.
        Then, increment t by multiplying it by the coarse increment.
        When relevant_thresh probability has entered the target state, step back to the previous coarse state, and
        swap over to incrementing with the fine increment.
        This allows you to efficiently sweep log-space.

        Procedurally, this starts probability in specified `initial_states` according to `initial_state_probs`, and then
        propagates that probability through the transition matrix.
        The FPT distribution is measured by tracking new probability entering the target state at each time.

        Note that absorbing boundary conditions are stripped from the transition matrix -- if this is not done, then
        the result is like a probability CDF, not a probability distribution.

        Parameters
        ----------
        Tmatrix: array-like
            Transition matrix

        initial_states: array-like of ints
            List of initial states to start probability in

        initial_state_probs: array-like
            Probability distribution across the initial states.

        target_states: array-like
            Target states for MFPT.

        tau
        increment: float
            Multiplicative increment for coarse steps
        fine_increment: float
            Multiplicative increment for fine steps, once the minimum probability in the target has been reached.
        relevant_thresh: float
            Amount of probability that must be in the target before switching to fine increments.
        max_steps: int
            Maximum number of steps to run
        max_time: float
            Maximum time to run to
        explicit_renormalization: bool
            Whether to explicitly renormalize the transition matrix. This should not be necessary -- if it is, there's
            probably some numerical instability you should be careful of.
        verbose: bool
            Produce verbose text output.

        Returns
        -------
        FPT distribution,
        probability distribution at each time,
        last step index,
        times at which FPT distribution was evaluated
        """

        n_states = len(Tmatrix)

        all_probabilities = np.full(shape=(max_steps + 1, n_states), fill_value=np.nan)

        # The initial probability vector is zero except in the origin states,
        #    which have their relative probabilities
        initial_probability = np.zeros(n_states)
        initial_probability[initial_states] = initial_state_probs
        initial_probability /= sum(initial_probability)

        all_probabilities[0] = initial_probability

        # Make the target states absorbing
        non_recycling_matrix = Tmatrix.copy()
        non_recycling_matrix[target_states, :] = 0.0
        for target in target_states:
            non_recycling_matrix[target, target] = 1.0

        # Track the probability that flowed into the target at each time
        probs = np.zeros(shape=max_steps)
        probs[0] = 0.0

        # At each one of our timesteps, track the amount of flux that entered the target
        last_step = 1

        get_next_step = lambda x: x * increment
        in_relevant_region = False

        steps = [1]

        with tqdm.tqdm(total=1) as pbar:

            for i in range(max_steps - 1):

                this_step = int(get_next_step(last_step))
                if this_step <= last_step:
                    this_step = int(last_step + 1)

                matrix_next = np.linalg.matrix_power(non_recycling_matrix, this_step)

                if explicit_renormalization:
                    matrix_next = matrix_next / np.sum(matrix_next, axis=1)

                probability = initial_probability @ matrix_next

                if explicit_renormalization:
                    probability /= sum(probability)

                # Check if we're just starting to get any probability
                if (
                    i > 0
                    and not in_relevant_region
                    and (sum(probability[target_states]) - sum(probs[: i + 1]))
                    > relevant_thresh
                ):
                    if verbose:
                        print(
                            f"*** Entered relevant region  at step {this_step}. "
                            f"Swapping to fine-grained, and taking a step back to {this_step / increment}."
                        )
                    # If so, then change our increment to finer resolution
                    # TODO: Would be cool to do something like as the probability increases,
                    #    continue scaling down to some minimum increment
                    in_relevant_region = True
                    this_step /= increment

                    steps.append(this_step)
                    all_probabilities[i + 1] = all_probabilities[i]
                    probs[i + 1] = probs[i]

                    get_next_step = lambda x: x * fine_increment

                    if verbose:
                        print(
                            f"Current time is {this_step}, time step will be {get_next_step(this_step)}"
                        )

                    continue

                steps.append(this_step)

                all_probabilities[i + 1] = probability

                # The amount that flowed INTO the target is the probability that's flowed in since the last t
                if i == 0:
                    # In the first iteration,  all the probability into the target just got there
                    probs[i + 1] = sum(probability[target_states])
                else:
                    # After the first, it's the amount that's there now minus the total amount that entered up until now
                    probs[i + 1] = sum(probability[target_states]) - sum(probs[: i + 1])

                pbar.update(probs[i + 1])

                # Check if we're done (i.e., all our probability has flowed  into  the target, none left.)
                if np.isclose(sum(probs), 1):
                    #         if np.isclose(probs[i+1], 1):
                    print(
                        f"*** All probability  reached the target at time {this_step}"
                    )
                    break

                if this_step > max_time:
                    print(
                        "*** Max steps reached, before all probability flowed into target."
                    )
                    break

                last_step = this_step

        print(f"Finished in {i} steps")
        print(
            f"By the last time, {sum(probs[:i])} probability has reached the target. (This should be 1!)"
        )

        times = np.array(steps, dtype=float) * float(tau)
        return probs[: i + 2], all_probabilities[: i + 2], i, times

    @classmethod
    def calc_fmatrix(
        cls,
        Fmatrix,
        tmatrix,
        prevFmatrix,
        list_of_pdfs,
        lag_list,
        ini_state,
        istateIndex,
        f_state,
    ):
        # Calculate FPT distribution from a the recursive formula, Eq. 3 in the paper below:
        # E. Suarez, A. J. Pratt, L. T. Chong, D. M. Zuckerman, Protein Science 26, 67-78 (2016).
        for time_index, time in enumerate(lag_list):
            # obtain the new transition matrix from time_index-1 to time_index
            if time_index == 0:
                tmatrix_new = np.linalg.matrix_power(tmatrix, time)
            else:
                tmatrix_new = np.linalg.matrix_power(
                    tmatrix, time - lag_list[time_index - 1]
                )
            Fmatrix = np.dot(tmatrix_new, prevFmatrix - np.diag(np.diag(prevFmatrix)))
            list_of_pdfs[istateIndex, time_index] = Fmatrix[
                ini_state[istateIndex], f_state
            ]
            prevFmatrix = Fmatrix


class MarkovFPT(MatrixFPT):
    """Derived a class for calculating FPTs using Markovian transition matrix"""

    @classmethod
    def mean_fpts(cls, markov_tmatrix, stateA, stateB, lag_time=1):
        """Computes mean first passage times using Markovian transition matrix

        in both directions A->B and B->A from a markov model. The MFPTs computed
        in this way are directly comparable with the values obtained by a long back
        and forth simulation between the target states.

        Parameters:
        -----------
        markov_matrix:      Numpy 2D array
                            Markovian transition matrix

        stateA, stateB:     List of integers
                            Both states are a list of indexes.

        lag_time:           integer
                            Lag time used, the trajectory is "observed" every lag_time
                            time steps

        Returns
        -------
        mean-first passage times from A->B and B->A

        """
        auxiliar_matrix = utils.pseudo_nm_tmatrix(markov_tmatrix, stateA, stateB)

        # Is going to return a Markovian mfpt since the auxiliar
        # matrix was build from a pure Markovian matrix
        return NonMarkovFPT.mean_fpts(auxiliar_matrix, stateA, stateB, lag_time)

    @classmethod
    def markov_commute_time(cls, transition_matrix, stateA, stateB, lag_time=1):
        """Computing commute time for Markovian Model

        Parameters:
        -----------
        transition_matrix:  Numpy 2D array
                            Markovian transition matrix

        stateA, stateB:     List of integers
                            Both states are a list of indexes.

        lag_time:           integer
                            Lag time used, the trajectory is "observed" every lag_time
                            time steps

        Returns
        -------
        Commute time from mean-first passage times
        """
        mfpts = cls.mean_fpts(transition_matrix, stateA, stateB, lag_time)
        return mfpts["mfptAB"] + mfpts["mfptBA"]


class NonMarkovFPT(MatrixFPT):
    """Derived a class for calculating FPTs using Non Markov transition matrix"""

    @classmethod
    def mean_fpts(cls, nm_transition_matrix, stateA, stateB, lag_time=1):
        """Computes the mean first passage times from a non-markovian model

        in both directions of A->B and B->A. The shape of the transition matrix
        should be (2*n_states, 2*n_states).
        Parameters:
        -----------
        nm_transition_matrix:   Numpy 2D array
                                Non-Markovian transition matrix

        stateA, stateB:         List of integers
                                Both states are a list of indexes.

        lag_time:               integer
                                Lag time used, the trajectory is "observed" every lag_time
                                time steps

        Returns
        -------
        mean-first passage times from A->B and B->A

        """
        utils.check_tmatrix(nm_transition_matrix)

        labeled_pops = utils.pops_from_tmatrix(nm_transition_matrix)

        n_states = len(labeled_pops) // 2

        fluxAB = 0
        fluxBA = 0

        for i in range(0, 2 * n_states, 2):
            for j in range(2 * n_states):
                if int(j / 2) in stateB:
                    fluxAB += labeled_pops[i] * nm_transition_matrix[i, j]

        for i in range(1, 2 * n_states + 1, 2):
            for j in range(2 * n_states):
                if int(j / 2) in stateA:
                    fluxBA += labeled_pops[i] * nm_transition_matrix[i, j]

        pop_colorA = 0.0
        pop_colorB = 0.0

        for i in range(0, 2 * n_states, 2):
            pop_colorA += labeled_pops[i]

        for i in range(1, 2 * n_states + 1, 2):
            pop_colorB += labeled_pops[i]

        if fluxAB == 0:
            mfptAB = float("inf")
        else:
            mfptAB = pop_colorA / fluxAB

        if fluxBA == 0:
            mfptBA = float("inf")
        else:
            mfptBA = pop_colorB / fluxBA

        mfptAB *= lag_time
        mfptBA *= lag_time

        return dict(mfptAB=mfptAB, mfptBA=mfptBA)



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

#with open("README.rst") as readme_file:
#    readme = readme_file.read()

#with open("HISTORY.rst") as history_file:
#    history = history_file.read()

# WESTPA-2.0 with westpa.analysis is not yet in Conda
#requirements = ["westpa>=v2.0b5"]
requirements = [
    "scikit-learn>=0.24,<1.1",
    "scipy>=1.5",
    "numpy>=1.16.5",
    "mdtraj>=1.9",
    "ray>=1.0",
    "h5py>=3.1",
    "tqdm",
    "rich",
    "toml"
]

setup_requirements = [
    "pytest-runner",
]

# TODO: Can maybe remove this?
test_requirements = ["pytest>=3"]

EXTRAS_REQUIRE = {
    "tests": ["pytest>=3", "pytest-timeout", "mdtraj"],
}

setup(
    author="John Russo",
    author_email="russojd@ohsu.edu",
    python_requires=">=3.5",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="Code for Markov state modeling of weighted ensemble trajectories.",
    entry_points={"console_scripts": ["msm_we=msm_we.cli:main"]},
    install_requires=requirements,
    license="MIT license",
#    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="msm_we",
    name="msm_we",
    packages=find_packages(include=["msm_we", "msm_we.*"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    extras_require=EXTRAS_REQUIRE,
    url="https://github.com/jdrusso/msm_we",
    version="0.1.13",
    zip_safe=False,
)
